/*****************************************************************************
 * Title:   GLBoing
 * Desc:    Tribute to Amiga Boing.
 * Author:  Jim Brooks  <gfx@jimbrooks.org>
 *          Original Amiga authors were R.J. Mical and Dale Luck.
 *          GLFW conversion by Marcus Geelnard
 * Notes:   - 360' = 2*PI [radian]
 *
 *          - Distances between objects are created by doing a relative
 *            Z translations.
 *
 *          - Although OpenGL enticingly supports alpha-blending,
 *            the shadow of the original Boing didn't affect the color
 *            of the grid.
 *
 *          - [Marcus] Changed timing scheme from interval driven to frame-
 *            time based animation steps (which results in much smoother
 *            movement)
 *
 * History of Amiga Boing:
 *
 * Boing was demonstrated on the prototype Amiga (codenamed "Lorraine") in
 * 1985. According to legend, it was written ad-hoc in one night by
 * R. J. Mical and Dale Luck. Because the bouncing ball animation was so fast
 * and smooth, attendees did not believe the Amiga prototype was really doing
 * the rendering. Suspecting a trick, they began looking around the booth for
 * a hidden computer or VCR.
 *****************************************************************************/

// Converted to MicroGlut by Ingemar
// FIRST VERSION USING MODERN OpenGL! 2021-05-02

#include <OpenGL/gl.h>
#include "MicroGlut.h"
// uses framework Cocoa
#include "VectorUtils3.h"
#include "GL_utilities.h"
#include "glugg.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define TRUE 1
#define FALSE 0

/*****************************************************************************
 * Various declarations and macros
 *****************************************************************************/

/* Prototypes */
void init( void );
void display( void );
void reshape( int w, int h );
void key_callback( unsigned char key, int x, int y);
void mouse_button_callback( int button, int state, int x, int y);
void cursor_position_callback( int x, int y );
GLuint BuildBoingBall(int *count, GLuint program);
void DrawBoingBall( void );
void BounceBall( double dt );
void BuildBoingBallBand( GLfloat long_lo, GLfloat long_hi );
GLuint CreateGrid(int *count, GLuint program);
void DrawGrid( void );

#define RADIUS           70.f
#define STEP_LONGITUDE   22.5f                   /* 22.5 makes 8 bands like original Boing */
#define STEP_LATITUDE    22.5f

#define DIST_BALL       (RADIUS * 2.f + RADIUS * 0.1f)

#define VIEW_SCENE_DIST (DIST_BALL * 3.f + 200.f)/* distance from viewer to middle of boing area */
#define GRID_SIZE       (RADIUS * 4.5f)          /* length (width) of grid */
#define BOUNCE_HEIGHT   (RADIUS * 2.1f)
#define BOUNCE_WIDTH    (RADIUS * 2.1f)

#define SHADOW_OFFSET_X -20.f
#define SHADOW_OFFSET_Y  10.f
#define SHADOW_OFFSET_Z   0.f

#define WALL_L_OFFSET   0.f
#define WALL_R_OFFSET   5.f

/* Animation speed (50.0 mimics the original GLUT demo speed) */
#define ANIMATION_SPEED 50.f

/* Maximum allowed delta time per physics iteration */
#define MAX_DELTA_T 0.02f

/* Draw ball, or its shadow */
typedef enum { DRAW_BALL, DRAW_BALL_SHADOW } DRAW_BALL_ENUM;

/* Global vars */
int windowed_xpos, windowed_ypos, windowed_width, windowed_height;
int width, height;
GLfloat deg_rot_y       = 0.f;
GLfloat deg_rot_y_inc   = 2.f;
int override_pos        = FALSE;
GLfloat cursor_x        = 0.f;
GLfloat cursor_y        = 0.f;
GLfloat ball_x          = -RADIUS;
GLfloat ball_y          = -RADIUS;
GLfloat ball_x_inc      = 1.f;
GLfloat ball_y_inc      = 2.f;
DRAW_BALL_ENUM drawBallHow;
double  t;
double  t_old = 0.f;
double  dt;

/* Random number generator */
#ifndef RAND_MAX
 #define RAND_MAX 4095
#endif

GLuint program; // Shader

/*****************************************************************************
 * Truncate a degree.
 *****************************************************************************/
GLfloat TruncateDeg( GLfloat deg )
{
   if ( deg >= 360.f )
      return (deg - 360.f);
   else
      return deg;
}

/*****************************************************************************
 * Convert a degree (360-based) into a radian.
 * 360' = 2 * PI
 *****************************************************************************/
double deg2rad( double deg )
{
   return deg / 360 * (2 * M_PI);
}

/*****************************************************************************
 * 360' sin().
 *****************************************************************************/
double sin_deg( double deg )
{
   return sin( deg2rad( deg ) );
}

/*****************************************************************************
 * 360' cos().
 *****************************************************************************/
double cos_deg( double deg )
{
   return cos( deg2rad( deg ) );
}

/*****************************************************************************
 * Compute a cross product (for a normal vector).
 *
 * c = a x b NO but n = (b-a)x(c-a)
 *****************************************************************************/
void CrossProductFromPoints( vec3 a, vec3 b, vec3 c, vec3 *n )
{
   GLfloat u1, u2, u3;
   GLfloat v1, v2, v3;

   u1 = b.x - a.x;
   u2 = b.y - a.y;
   u3 = b.y - a.z;

   v1 = c.x - a.x;
   v2 = c.y - a.y;
   v3 = c.z - a.z;

   n->x = u2 * v3 - v2 * u3;
   n->y = u3 * v1 - v3 * u1;
   n->z = u1 * v2 - v1 * u2;
}


#define BOING_DEBUG 0

GLuint ballID;
int ballCount;
GLuint gridID;
int gridCount;
mat4 projection, view;

/*****************************************************************************
 * init()
 *****************************************************************************/
void init( void )
{
   /*
    * Background.
    */
   glClearColor( 0.55f, 0.55f, 0.55f, 0.f );

	program = loadShaders("boing.vert", "boing.frag");

	ballID = BuildBoingBall(&ballCount, program);
	gridID = CreateGrid(&gridCount, program);

   projection = perspective(2.f * (float) atan2( RADIUS, 200.f ),
                       1.0,
                       1.f, VIEW_SCENE_DIST);
                       
   projection = frustum(-1, 1, -1, 1, 1, 200);

    glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, &projection.m);
}

/*****************************************************************************
 * display()
 *****************************************************************************/
void display(void)
{
   /* Timing */
    t = glutGet(GLUT_ELAPSED_TIME) / 1000.0;
    dt = t - t_old;
    t_old = t;

   glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

   drawBallHow = DRAW_BALL_SHADOW;
   DrawBoingBall();

   DrawGrid();

   drawBallHow = DRAW_BALL;
   DrawBoingBall();

   glutSwapBuffers();
}

/*****************************************************************************
 * reshape()
 *****************************************************************************/
void reshape( int w, int h )
{
   glViewport( 0, 0, (GLsizei)w, (GLsizei)h );

//   projection = perspective(2.f * (float) atan2( RADIUS, 200.f ),
//                       (float)w / (float)h,
//                       1.f, VIEW_SCENE_DIST);
    projection = frustum(-0.35, 0.35, -0.35, 0.35, 1, 1000);
    glUniformMatrix4fv(glGetUniformLocation(program, "projectionMatrix"), 1, GL_TRUE, &projection.m);

    vec3 eye = { 0.f, 0.f, VIEW_SCENE_DIST };
    vec3 center = { 0.f, 0.f, 0.f };
    vec3 up = { 0.f, -1.f, 0.f };
    mat4 view = lookAtv(eye, center, up);
    glUniformMatrix4fv(glGetUniformLocation(program, "worldViewMatrix"), 1, GL_TRUE, &view.m);
}

void key_callback( unsigned char key, int x, int y)
{
    if (key == GLUT_KEY_ESC)
        glutExit();
}

static void set_ball_pos ( GLfloat x, GLfloat y )
{
   ball_x = (width / 2) - x;
   ball_y = y - (height / 2);
}

void mouse_button_callback( int button, int state, int x, int y )
{
   if (button != GLUT_LEFT_BUTTON)
      return;

   if (state == GLUT_DOWN)
   {
      override_pos = TRUE;
      set_ball_pos(cursor_x, cursor_y);
   }
   else
   {
      override_pos = FALSE;
   }
}

void cursor_position_callback( int x, int y )
{
   cursor_x = (float) x;
   cursor_y = (float) y;

   if ( override_pos )
      set_ball_pos(cursor_x, cursor_y);
}

/*****************************************************************************
 * Draw the Boing ball.
 *
 * The Boing ball is sphere in which each facet is a rectangle.
 * Facet colors alternate between red and white.
 * The ball is built by stacking latitudinal circles.  Each circle is composed
 * of a widely-separated set of points, so that each facet is noticeably large.
 *****************************************************************************/
void DrawBoingBall( void )
{
	// Location, shader, draw
	
	GLfloat lon_deg;     /* degree of longitude */
  	double dt_total, dt2;
  	mat4 m;

  /*
   * Another relative Z translation to separate objects.
   */
   m = T( 0.0, 0.0, DIST_BALL );

   /* Update ball position and rotation (iterate if necessary) */
   dt_total = dt;
   while( dt_total > 0.0 )
   {
       dt2 = dt_total > MAX_DELTA_T ? MAX_DELTA_T : dt_total;
       dt_total -= dt2;
       BounceBall( dt2 );
       deg_rot_y = TruncateDeg( deg_rot_y + deg_rot_y_inc*((float)dt2*ANIMATION_SPEED) );
   }

   /* Set ball position */
   m = Mult(m, T( ball_x, ball_y, 0.0 ));

  /*
   * Offset the shadow.
   */
   if ( drawBallHow == DRAW_BALL_SHADOW )
   {
      m = Mult(
      m, T( SHADOW_OFFSET_X,
                    SHADOW_OFFSET_Y,
                    SHADOW_OFFSET_Z ));
   }

  /*
   * Tilt the ball.
   */
   m = Mult( m, Rz( -20.0 *M_PI/180.0));

  /*
   * Continually rotate ball around Y axis.
   */
   m = Mult( m, Ry( deg_rot_y *M_PI/180.0 ));

  /*
   * Set OpenGL state for Boing ball.
   */
   glCullFace( GL_FRONT );
   glEnable( GL_CULL_FACE );

    glUniform1i(glGetUniformLocation(program, "drawBallHow"), drawBallHow);

    glUniformMatrix4fv(glGetUniformLocation(program, "modelWorldMatrix"), 1, GL_TRUE, &m.m);
	glBindVertexArray(ballID);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, ballCount);
}



// Build and upload geometry
GLuint BuildBoingBall(int *count, GLuint program)
{
   GLfloat lon_deg;     /* degree of longitude */
   double dt_total, dt2;

  /*
   * Build a faceted latitude slice of the Boing ball,
   * stepping same-sized vertical bands of the sphere.
   */
   gluggBegin(GLUGG_QUADS);
   for ( lon_deg = 0;
         lon_deg < 180;
         lon_deg += STEP_LONGITUDE )
   {
     /*
      * Draw a latitude circle at this longitude.
      */
      BuildBoingBallBand( lon_deg,
                         lon_deg + STEP_LONGITUDE );
   }
   return gluggEnd(count, program, 0);
}


/*****************************************************************************
 * Bounce the ball.
 *****************************************************************************/
void BounceBall( double delta_t )
{
   GLfloat sign;
   GLfloat deg;

   if ( override_pos )
     return;

   /* Bounce on walls */
   if ( ball_x >  (BOUNCE_WIDTH/2 + WALL_R_OFFSET ) )
   {
      ball_x_inc = -0.5f - 0.75f * (GLfloat)rand() / (GLfloat)RAND_MAX;
      deg_rot_y_inc = -deg_rot_y_inc;
   }
   if ( ball_x < -(BOUNCE_HEIGHT/2 + WALL_L_OFFSET) )
   {
      ball_x_inc =  0.5f + 0.75f * (GLfloat)rand() / (GLfloat)RAND_MAX;
      deg_rot_y_inc = -deg_rot_y_inc;
   }

   /* Bounce on floor / roof */
   if ( ball_y >  BOUNCE_HEIGHT/2      )
   {
      ball_y_inc = -0.75f - 1.f * (GLfloat)rand() / (GLfloat)RAND_MAX;
   }
   if ( ball_y < -BOUNCE_HEIGHT/2*0.85 )
   {
      ball_y_inc =  0.75f + 1.f * (GLfloat)rand() / (GLfloat)RAND_MAX;
   }

   /* Update ball position */
   ball_x += ball_x_inc * ((float)delta_t*ANIMATION_SPEED);
   ball_y += ball_y_inc * ((float)delta_t*ANIMATION_SPEED);

  /*
   * Simulate the effects of gravity on Y movement.
   */
   if ( ball_y_inc < 0 ) sign = -1.0; else sign = 1.0;

   deg = (ball_y + BOUNCE_HEIGHT/2) * 90 / BOUNCE_HEIGHT;
   if ( deg > 80 ) deg = 80;
   if ( deg < 10 ) deg = 10;

   ball_y_inc = sign * 4.f * (float) sin_deg( deg );
}


/*****************************************************************************
 * Draw a faceted latitude band of the Boing ball.
 *
 * Parms:   long_lo, long_hi
 *          Low and high longitudes of slice, resp.
 *****************************************************************************/

void BuildBoingBallBand( GLfloat long_lo,
                        GLfloat long_hi )
{
   vec3 vert_ne;            /* "ne" means south-east, so on */
   vec3 vert_nw;
   vec3 vert_sw;
   vec3 vert_se;
   vec3 vert_norm;
   GLfloat  lat_deg;
   static int colorToggle = 0;

  /*
   * Iterate through the points of a latitude circle.
   * A latitude circle is a 2D set of X,Z points.
   */
   for ( lat_deg = 0;
         lat_deg <= (360 - STEP_LATITUDE);
         lat_deg += STEP_LATITUDE )
   {
     /*
      * Color this polygon with red or white.
      * NOTE: GLUGG does not support colors, so I pass the information
      * as texture coordinates instead, to be turned to color in the shader.
      */
      if ( colorToggle )
      	gluggTexCoord(0.8, 0.1); // will cause red in shader
      else
      	gluggTexCoord(0.95, 0.95); // will cause white in shader
#if 0
      if ( lat_deg >= 180 )
         if ( colorToggle )
            glColor3f( 0.1f, 0.8f, 0.1f );
         else
            glColor3f( 0.5f, 0.5f, 0.95f );
#endif
      colorToggle = ! colorToggle;

     /*
      * Change color if drawing shadow.
      */
      if ( drawBallHow == DRAW_BALL_SHADOW )
         gluggTexCoord(0.35, 0.35); // Use texture coordinates to pass color

     /*
      * Assign each Y.
      */
      vert_ne.y = vert_nw.y = (float) cos_deg(long_hi) * RADIUS;
      vert_sw.y = vert_se.y = (float) cos_deg(long_lo) * RADIUS;

     /*
      * Assign each X,Z with sin,cos values scaled by latitude radius indexed by longitude.
      * Eg, long=0 and long=180 are at the poles, so zero scale is sin(longitude),
      * while long=90 (sin(90)=1) is at equator.
      */
      vert_ne.x = (float) cos_deg( lat_deg                 ) * (RADIUS * (float) sin_deg( long_lo + STEP_LONGITUDE ));
      vert_se.x = (float) cos_deg( lat_deg                 ) * (RADIUS * (float) sin_deg( long_lo                  ));
      vert_nw.x = (float) cos_deg( lat_deg + STEP_LATITUDE ) * (RADIUS * (float) sin_deg( long_lo + STEP_LONGITUDE ));
      vert_sw.x = (float) cos_deg( lat_deg + STEP_LATITUDE ) * (RADIUS * (float) sin_deg( long_lo                  ));

      vert_ne.z = (float) sin_deg( lat_deg                 ) * (RADIUS * (float) sin_deg( long_lo + STEP_LONGITUDE ));
      vert_se.z = (float) sin_deg( lat_deg                 ) * (RADIUS * (float) sin_deg( long_lo                  ));
      vert_nw.z = (float) sin_deg( lat_deg + STEP_LATITUDE ) * (RADIUS * (float) sin_deg( long_lo + STEP_LONGITUDE ));
      vert_sw.z = (float) sin_deg( lat_deg + STEP_LATITUDE ) * (RADIUS * (float) sin_deg( long_lo                  ));

     /*
      * Draw the facet.
      */

      CrossProductFromPoints( vert_ne, vert_nw, vert_sw, &vert_norm );
      gluggNormal( vert_norm.x, vert_norm.y, vert_norm.z );

      gluggVertex( vert_ne.x, vert_ne.y, vert_ne.z );
      gluggVertex( vert_nw.x, vert_nw.y, vert_nw.z );
      gluggVertex( vert_sw.x, vert_sw.y, vert_sw.z );
      gluggVertex( vert_se.x, vert_se.y, vert_se.z );

#if BOING_DEBUG
      printf( "----------------------------------------------------------- \n" );
      printf( "lat = %f  long_lo = %f  long_hi = %f \n", lat_deg, long_lo, long_hi );
      printf( "vert_ne  x = %.8f  y = %.8f  z = %.8f \n", vert_ne.x, vert_ne.y, vert_ne.z );
      printf( "vert_nw  x = %.8f  y = %.8f  z = %.8f \n", vert_nw.x, vert_nw.y, vert_nw.z );
      printf( "vert_se  x = %.8f  y = %.8f  z = %.8f \n", vert_se.x, vert_se.y, vert_se.z );
      printf( "vert_sw  x = %.8f  y = %.8f  z = %.8f \n", vert_sw.x, vert_sw.y, vert_sw.z );
#endif

   }

  /*
   * Toggle color so that next band will opposite red/white colors than this one.
   */
   colorToggle = ! colorToggle;

  /*
   * This circular band is done.
   */
   return;
}


/*****************************************************************************
 * Draw the purple grid of lines, behind the Boing ball.
 * When the Workbench is dropped to the bottom, Boing shows 12 rows.
 *****************************************************************************/

void DrawGrid( void )
{
   glDisable( GL_CULL_FACE );

  /*
   * Another relative Z translation to separate objects.
   */
   
   mat4 m;
   m = T( 0.0, 0.0, DIST_BALL );

    glUniform1i(glGetUniformLocation(program, "drawBallHow"), DRAW_BALL); // Means color normally

// DRAW THE GRID HERE
    glUniformMatrix4fv(glGetUniformLocation(program, "modelWorldMatrix"), 1, GL_TRUE, &m.m);
	glBindVertexArray(gridID);	// Select VAO
	glDrawArrays(GL_TRIANGLES, 0, gridCount);
}


GLuint CreateGrid(int *count, GLuint program)
{
   int              row, col;
   const int        rowTotal    = 12;                   /* must be divisible by 2 */
   const int        colTotal    = rowTotal;             /* must be same as rowTotal */
   const GLfloat    widthLine   = 2.0;                  /* should be divisible by 2 */
   const GLfloat    sizeCell    = GRID_SIZE / rowTotal;
   const GLfloat    z_offset    = -40.0;
   GLfloat          xl, xr;
   GLfloat          yt, yb;

	gluggBegin(GLUGG_QUADS);
  /*
   * Draw vertical lines (as skinny 3D rectangles).
   */
   for ( col = 0; col <= colTotal; col++ )
   {
     /*
      * Compute co-ords of line.
      */
      xl = -GRID_SIZE / 2 + col * sizeCell;
      xr = xl + widthLine;

      yt =  GRID_SIZE / 2;
      yb = -GRID_SIZE / 2 - widthLine;

      gluggTexCoord( 0.6f, 0.1f /*, 0.6f*/ );               /* purple */

// Draw quad as two triangles

      gluggVertex( xr, yt, z_offset );       /* NE */
      gluggVertex( xl, yt, z_offset );       /* NW */
      gluggVertex( xl, yb, z_offset );       /* SW */
      gluggVertex( xr, yb, z_offset );       /* SE */
   }

  /*
   * Draw horizontal lines (as skinny 3D rectangles).
   */
   for ( row = 0; row <= rowTotal; row++ )
   {
     /*
      * Compute co-ords of line.
      */
      yt = GRID_SIZE / 2 - row * sizeCell;
      yb = yt - widthLine;

      xl = -GRID_SIZE / 2;
      xr =  GRID_SIZE / 2 + widthLine;

      gluggTexCoord( 0.6f, 0.1f  /*, 0.6f */ );               /* purple */

// Draw quad as two triangles
      gluggVertex( xr, yt, z_offset );       /* NE */
      gluggVertex( xl, yt, z_offset );       /* NW */
      gluggVertex( xl, yb, z_offset );       /* SW */
      gluggVertex( xr, yb, z_offset );       /* SE */
   }

   return gluggEnd(count, program, 0);
}


/*======================================================================*
 * main()
 *======================================================================*/

int main(int argc, char *argv[])
{
	glutInit(&argc, argv);
	glutInitContextVersion(3, 2);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowSize(400, 400);
	glutCreateWindow ("Boing (classic Amiga demo)");
	glutDisplayFunc(display);
	glutKeyboardFunc(key_callback);
	glutPassiveMotionFunc(cursor_position_callback);
	glutMouseFunc(mouse_button_callback);
	glutRepeatingTimer(15);
	glutReshapeFunc(reshape);
	init ();
	glutMainLoop();
	return 0;
}

